""" Tests for system tools """

import os
import subprocess
import pytest
from ubiquerg import is_command_callable
from veracitools import ExpectContext

__author__ = "Vince Reuter"
__email__ = "vreuter@virginia.edu"

EXTENSIONS = [
    ".py",
    ".rb",
    ".rs",
    ".sh",
    ".java",
    ".jar",
    ".pl",
    ".o",
    ".R",
    ".r",
    ".cpp",
    ".c",
    ".hs",
    ".scala",
    ".class",
]


def _mkfile(f):
    """'touch' a file."""
    with open(f, "w"):
        print("Touch: {}".format(f))
    return f


@pytest.mark.parametrize(
    ["cmd", "exp"],
    [("man", True), ("not-a-cmd", False), (None, TypeError), ("", ValueError)],
)
def test_command_callability_check_basic(cmd, exp):
    """Verify expected behavior of command callability checker."""
    with ExpectContext(exp, is_command_callable) as check_callable:
        check_callable(cmd)


@pytest.mark.parametrize(
    "relpath",
    ["subdir", os.path.join("subA", "subB")] + ["testfile" + ext for ext in EXTENSIONS],
)
def test_missing_path_isnt_callable(tmpdir, relpath):
    """A filepath that doesn't exist (and isn't on PATH) isn't callable."""
    p = os.path.join(tmpdir.strpath, relpath)
    assert not os.path.exists(p)
    assert not is_command_callable(p)


@pytest.mark.parametrize("get_arg", [lambda d: d, lambda d: os.path.join(d, "subdir")])
def test_extant_folder_isnt_callable(tmpdir, get_arg):
    """A directory path can't be callable."""
    p = get_arg(tmpdir.strpath)
    print("p: {}".format(p))
    if not os.path.exists(p):
        os.makedirs(p)
    assert os.path.exists(p)
    assert not is_command_callable(p)


@pytest.mark.parametrize(
    "filepath", [lambda d: os.path.join(d, "testfile" + ext) for ext in EXTENSIONS]
)
@pytest.mark.parametrize(
    ["setup", "exp_exe"],
    [
        (lambda _: None, False),
        (lambda f: subprocess.check_call(["chmod", "+x", f]), True),
    ],
)
def test_extant_file_requires_exec_for_callability(tmpdir, filepath, setup, exp_exe):
    """Filepath is callable iff it has exec bit."""
    f = filepath(tmpdir.strpath)
    assert not os.path.exists(f)
    _mkfile(f)
    assert os.path.isfile(f)
    setup(f)
    assert os.access(f, os.X_OK) is exp_exe
    assert is_command_callable(f) is exp_exe



""" Tests for web """

import pytest
from ubiquerg.web import is_url


@pytest.mark.parametrize("s", ["https://www.github.com", "https://www.youtube.com"])
def test_is_url_tests_positive(s):
    assert is_url(s)


@pytest.mark.parametrize(
    "s", ["www.github.com", "test: string spaces", "j%2vv@::https://test.com"]
)
def test_is_url_tests_negative(s):
    assert not is_url(s)



""" Tests for rendering CLI options and arguments """

from collections import OrderedDict
import pytest
from ubiquerg import build_cli_extra, powerset, VersionInHelpParser, convert_value
import argparse


def pytest_generate_tests(metafunc):
    """Test case generation and parameterization for this module"""
    if "ordwrap" in metafunc.fixturenames:
        metafunc.parametrize("ordwrap", [tuple, OrderedDict])


def build_parser():
    """
    Example argument parser, needed solely for testing purposes.
    This example parser was copied from looper/__init__.py

    :return argparse.ArgumentParser
    """

    banner = "%(prog)s - Loop through samples and submit pipelines."
    additional_description = (
        "For subcommand-specific options, " "type: '%(prog)s <subcommand> -h'"
    )
    additional_description += "\nhttps://github.com/pepkit/looper"

    parser = VersionInHelpParser(
        prog="looper", description=banner, epilog=additional_description, version=0.1
    )

    # Logging control
    parser.add_argument(
        "--logfile",
        dest="logfile",
        help="Optional output file for looper logs (default: %(default)s)",
    )
    parser.add_argument(
        "--verbosity",
        dest="verbosity",
        type=int,
        choices=range(4),
        help="Choose level of verbosity (default: %(default)s)",
    )
    parser.add_argument("--logging-level", dest="logging_level", help=argparse.SUPPRESS)
    parser.add_argument(
        "--dbg",
        dest="dbg",
        action="store_true",
        help="Turn on debug mode (default: %(default)s)",
    )
    parser.add_argument(
        "--env",
        dest="env",
        default=None,
        help="Environment variable that points to the DIVCFG file. "
        "(default: DIVCFG)",
    )
    parser.add_argument(
        "--dotfile-template",
        action="store_true",
        help="Print out a looper dotfile template and exit",
    )

    # Individual subcommands
    msg_by_cmd = {
        "run": "Main Looper function: Submit jobs for samples.",
        "rerun": "Resubmit jobs with failed flags.",
        "runp": "Submit jobs for a project.",
        "summarize": "Summarize statistics of project samples.",
        "destroy": "Remove all files of the project.",
        "check": "Checks flag status of current runs.",
        "clean": "Runs clean scripts to remove intermediate "
        "files of already processed jobs.",
    }

    subparsers = parser.add_subparsers(dest="command")

    def add_subparser(cmd):
        message = msg_by_cmd[cmd]
        return subparsers.add_parser(cmd, description=message, help=message)

    # Run and rerun command
    run_subparser = add_subparser("run")
    rerun_subparser = add_subparser("rerun")
    collate_subparser = add_subparser("runp")
    for subparser in [run_subparser, rerun_subparser, collate_subparser]:
        subparser.add_argument(
            "--ignore-flags",
            dest="ignore_flags",
            default=False,
            action="store_true",
            help="Ignore run status flags? Default: False. "
            "By default, pipelines will not be submitted if a pypiper "
            "flag file exists marking the run (e.g. as "
            "'running' or 'failed'). Set this option to ignore flags "
            "and submit the runs anyway. Default=False",
        )
        subparser.add_argument(
            "-t",
            "--time-delay",
            dest="time_delay",
            type=int,
            default=0,
            help="Time delay in seconds between job submissions.",
        )
        subparser.add_argument(
            "-p", "--package", help="Name of computing resource package to use"
        )
        subparser.add_argument(
            "--compute",
            help="Specification of individual computing resource settings; "
            "separate setting name/key from value with equals sign, "
            "and separate key-value pairs from each other by comma",
        )
        subparser.add_argument(
            "--limit", dest="limit", default=None, type=int, help="Limit to n samples."
        )
        subparser.add_argument(
            "-a",
            "--pipeline-args",
            dest="pipeline_args",
            default="",
            help="arguments to pass to a pipline",
        )
        subparser.add_argument(
            "-s",
            "--settings",
            dest="settings",
            default="",
            help="path to a YAML-formatted settings file used to populate "
            "the command template",
        )
    for subparser in [run_subparser, rerun_subparser]:
        # Note that defaults for otherwise numeric lump parameters are
        # set to
        # null by default so that the logic that parses their values may
        # distinguish between explicit 0 and lack of specification.
        subparser.add_argument(
            "--lump",
            default=None,
            type=float,
            help="Maximum total input file size for a lump/batch of "
            "commands in a single job (in GB)",
        )
        subparser.add_argument(
            "--lumpn",
            default=None,
            type=int,
            help="Number of individual scripts grouped into " "single submission",
        )

    # Other commands
    summarize_subparser = add_subparser("summarize")
    destroy_subparser = add_subparser("destroy")
    check_subparser = add_subparser("check")
    clean_subparser = add_subparser("clean")

    check_subparser.add_argument(
        "-A",
        "--all-folders",
        action="store_true",
        default=False,
        help="Check status for all project's output folders, not just "
        "those for samples specified in the config file used. "
        "Default=False",
    )

    for subparser in [destroy_subparser, clean_subparser]:
        subparser.add_argument(
            "--force-yes",
            action="store_true",
            default=False,
            help="Provide upfront confirmation of destruction intent, "
            "to skip console query.  Default=False",
        )

    # Common arguments
    for subparser in [
        run_subparser,
        rerun_subparser,
        summarize_subparser,
        destroy_subparser,
        check_subparser,
        clean_subparser,
        collate_subparser,
    ]:
        subparser.add_argument(
            "config_file", nargs="?", help="Project configuration file (YAML)."
        )
        # subparser.add_argument(
        #         "-c", "--config", required=False, default=None,
        #         dest="looper_config", help="Looper configuration file (
        #         YAML).")
        subparser.add_argument(
            "--pipeline-interfaces",
            dest="pifaces",
            nargs="+",
            action="append",
            help="Path to a pipeline interface file",
        )
        subparser.add_argument(
            "--file-checks",
            dest="file_checks",
            action="store_true",
            help="Perform input file checks. Default=True.",
        )
        subparser.add_argument(
            "-d",
            "--dry-run",
            dest="dry_run",
            action="store_true",
            help="Don't actually submit the jobs.  Default=False",
        )

        fetch_samples_group = subparser.add_argument_group(
            "select samples",
            "This group of arguments lets you specify samples to use by "
            "exclusion OR inclusion of the samples attribute values.",
        )
        fetch_samples_group.add_argument(
            "--selector-attribute",
            dest="selector_attribute",
            default="toggle",
            help="Specify the attribute for samples exclusion OR inclusion",
        )
        protocols = fetch_samples_group.add_mutually_exclusive_group()
        protocols.add_argument(
            "--selector-exclude",
            nargs="*",
            dest="selector_exclude",
            help="Operate only on samples that either lack this attribute "
            "value or for which this value is not in this collection.",
        )
        protocols.add_argument(
            "--selector-include",
            nargs="*",
            dest="selector_include",
            help="Operate only on samples associated with these attribute "
            "values; if not provided, all samples are used.",
        )
        subparser.add_argument(
            "--amendments",
            dest="amendments",
            nargs="+",
            help="Name of amendment(s) to use, as designated in the "
            "project's configuration file",
        )

    return parser, msg_by_cmd


@pytest.mark.parametrize(
    ["optargs", "expected"],
    [
        (
            [
                ("-X", None),
                ("--revert", 1),
                ("-O", "outfile"),
                ("--execute-locally", None),
                ("-I", ["file1", "file2"]),
            ],
            "-X --revert 1 -O outfile --execute-locally -I file1 file2",
        )
    ],
)
def test_build_cli_extra(optargs, expected, ordwrap):
    """Check that CLI optargs are rendered as expected."""
    observed = build_cli_extra(ordwrap(optargs))
    print("expected: {}".format(expected))
    print("observed: {}".format(observed))
    assert expected == observed


@pytest.mark.parametrize("optargs", powerset([(None, "a"), (1, "one")], nonempty=True))
def test_illegal_cli_extra_input_is_exceptional(optargs, ordwrap):
    """Non-string keys are illegal and cause a TypeError."""
    with pytest.raises(TypeError):
        build_cli_extra(ordwrap(optargs))


def test_dests_by_subparser_return_type():
    """Check if the return type is dict of lists keyed by subcommand name"""
    parser, msgs_by_cmd = build_parser()
    assert isinstance(parser.dests_by_subparser(), dict)
    assert all([isinstance(v, list) for k, v in parser.dests_by_subparser().items()])
    assert all([k in parser.dests_by_subparser() for k in list(msgs_by_cmd.keys())])


@pytest.mark.parametrize("subcmd", list(build_parser()[0].dests_by_subparser().keys()))
def test_dests_by_subparser_specific_subcmd(subcmd):
    """
    Should return only specified section of the
    return if subparser specified
    """
    parser, _ = build_parser()
    assert isinstance(parser.dests_by_subparser(subcmd), dict)
    assert subcmd in parser.dests_by_subparser(subcmd).keys()
    assert isinstance(parser.dests_by_subparser(subcmd)[subcmd], list)


def test_dests_by_subparser_toplevel_actions():
    """
    Should return only top level dests, which is a list
    """
    parser, _ = build_parser()
    tld = parser.dests_by_subparser(top_level=True)
    assert isinstance(tld, list)
    assert len(tld) == 6


def test_arg_defaults_return_type():
    """Check if the return type is dict of dicts keyed by subcommand name"""
    parser, msgs_by_cmd = build_parser()
    assert isinstance(parser.arg_defaults(), dict)
    assert all([isinstance(v, dict) for k, v in parser.arg_defaults().items()])
    assert all([k in parser.arg_defaults() for k in list(msgs_by_cmd.keys())])


@pytest.mark.parametrize("subcmd", list(build_parser()[0].dests_by_subparser().keys()))
def test_arg_defaults_specific_subcmd(subcmd):
    """
    Should return only specified section of the
    return if subparser specified
    """
    parser, _ = build_parser()
    assert isinstance(parser.arg_defaults(subcmd), dict)
    assert subcmd in parser.arg_defaults(subcmd).keys()
    assert isinstance(parser.arg_defaults(subcmd)[subcmd], dict)


def test_arg_defaults_toplevel_actions():
    """
    Should return only top level defaults, which is a dict
    """
    parser, _ = build_parser()
    tld = parser.arg_defaults(top_level=True)
    assert isinstance(tld, dict)
    assert len(tld) == 6


def test_arg_defaults_unqiue():
    """
    Should return only dict of unique kv pairs, not keyed by subcommands
    """
    parser, _ = build_parser()
    tld = parser.arg_defaults(unique=True)
    assert isinstance(tld, dict)
    assert len(tld) == 17
    assert len(set(tld.keys())) == len(tld.keys())


@pytest.mark.parametrize(
    ["input", "output"],
    [
        (None, None),
        ("None", None),
        ("1", 1),
        ("0.1", 0.1),
        ("True", True),
        ("False", False),
    ],
)
def test_value_conversion(input, output):
    assert convert_value(input) == output



import errno
import glob
import logging
import os

from contextlib import contextmanager
from .files import wait_for_lock, create_file_racefree
from signal import signal, SIGINT, SIGTERM
from typing import Union, List

PID = os.getpid()
READ = f"read-{PID}"
READ_GLOB = "read-*"
WRITE = "write"
UNIVERSAL = "universal"
LOCK_PREFIX = "lock"

_LOGGER = logging.getLogger(__name__)


class ThreeLocker(object):
    """
    A class to lock files for reading and writing.

    It uses a three-lock system, with separate read-lock, write-lock, and
    universal-lock (or lock-lock). The universal lock is used to lock the locks,
    to prevent race conditions between read and write locks. It allows multiple
    simultaneous readers, as long as there is no writer.
    It creates lock files in the same directory as the file to be locked.
    """

    def __init__(
        self, filepath: str, wait_max: int = 10, strict_ro_locks: bool = False
    ):
        self.wait_max = wait_max
        self.strict_ro_locks = strict_ro_locks
        self.set_file_path(filepath)
        self.locked = {READ: False, WRITE: False}

    @property
    def filepath(self) -> str:
        return self._filepath

    def set_file_path(self, filepath: str) -> str:
        if filepath:
            self._filepath = mkabs(filepath)
            self.lock_paths = make_all_lock_paths(self.filepath)
        else:
            self._filepath = None
            self.lock_paths = None

        return self._filepath

    def read_lock(self) -> bool:
        if not self.filepath:
            _LOGGER.warning("No filepath, no need to lock.")
            return True
        lock_path = self.lock_paths[READ]
        if not ensure_write_access(lock_path, self.strict_ro_locks):
            return True

        self.create_read_lock(self.filepath, self.wait_max)
        self.locked[READ] = True
        return True

    def write_lock(self) -> bool:
        if not self.filepath:
            _LOGGER.warning("No filepath, no need to lock.")
            return True

        lock_path = self.lock_paths[WRITE]
        if not ensure_write_access(lock_path, self.strict_ro_locks):
            # for writing, just fail anyway
            raise OSError(f"No write access to '{lock_path}'; can't lock file.")
        self.create_write_lock(self.filepath, self.wait_max)
        self.locked[READ] = True
        self.locked[WRITE] = True
        return True

    def read_unlock(self) -> bool:
        if not self.filepath:
            _LOGGER.warning("No filepath, no need to unlock.")
            return True
        _remove_lock(self.lock_paths[READ])
        self.locked[READ] = False
        return True

    def write_unlock(self) -> bool:
        if not self.filepath:
            _LOGGER.warning("No filepath, no need to unlock.")
            return True
        _remove_lock(self.lock_paths[WRITE])
        _remove_lock(self.lock_paths[READ])
        self.locked[WRITE] = False
        self.locked[READ] = False
        return True

    def create_read_lock(self, filepath: str = None, wait_max: int = None) -> None:
        """
        Securely create a read lock file

        :param str filepath: path to a file to lock
        :param int wait_max: max wait time if the file in question is already locked
        """
        filepath = filepath or self.filepath
        wait_max = wait_max or self.wait_max
        wait_for_lock(self.lock_paths[UNIVERSAL], wait_max)
        _create_lock(self.lock_paths[UNIVERSAL], filepath, wait_max)
        wait_for_lock(self.lock_paths[WRITE], wait_max)
        _create_lock(self.lock_paths[READ], filepath, wait_max)
        _remove_lock(self.lock_paths[UNIVERSAL])

    def create_write_lock(self, filepath: str = None, wait_max: int = None) -> None:
        """
        Securely create a write lock file

        :param str filepath: path to a file to lock
        :param int wait_max: max wait time if the file in question is already locked
        """
        filepath = filepath or self.filepath
        wait_max = wait_max or self.wait_max
        wait_for_lock(self.lock_paths[UNIVERSAL], wait_max)
        _create_lock(self.lock_paths[UNIVERSAL], filepath, wait_max)
        read_lock_paths = glob.glob(
            self.lock_paths[READ_GLOB]
        )  # must occur after universal lock is set
        all_lock_paths = read_lock_paths + [self.lock_paths[WRITE]]
        wait_for_locks(all_lock_paths, wait_max)
        _create_lock(self.lock_paths[READ], filepath, wait_max)
        _create_lock(self.lock_paths[WRITE], filepath, wait_max)
        _remove_lock(self.lock_paths[UNIVERSAL])

    def _interrupt_handler(self, signal_received, frame):
        if signal_received == SIGINT:
            _LOGGER.warning(f"Received SIGINT, unlocking file and exiting...")
            self.write_unlock()
            self.read_unlock()
            raise SystemExit
        if signal_received == SIGTERM:
            _LOGGER.warning(f"Received SIGTERM, unlocking file and exiting...")
            self.__exit__(None, None, None)
            self.write_unlock()
            self.read_unlock()
            raise SystemExit

    def __repr__(self) -> str:
        settings_dict = {
            "filepath": self.filepath,
            "wait_max": self.wait_max,
            "locked": self.locked,
            "strict_ro_locks": self.strict_ro_locks,
        }

        return f"{type(self).__name__}({settings_dict})"

    def __del__(self) -> None:
        if self.filepath:
            if self.locked[WRITE]:
                self.write_unlock()
            if self.locked[READ]:
                self.read_unlock()


def ensure_locked(type: str = WRITE):  # decorator factory
    """Decorator to apply to functions to make sure they only happen when locked."""

    def decorator(func):
        def inner_func(self, *args, **kwargs):
            if not self.locker:
                raise OSError("File not lockable. File locker not provided.")
            if not self.locker.locked[type]:
                raise OSError(
                    f"This function must use a context manager to {type}-lock the file"
                )

            return func(self, *args, **kwargs)

        return inner_func

    return decorator


@contextmanager
def read_lock(obj: Union[str, object]) -> object:
    """
    Read-lock a filepath or object with locker attribute.
    """
    if type(obj) == str:
        locker = ThreeLocker(obj)
    elif hasattr(obj, "locker"):
        locker = obj.locker
    else:
        raise AttributeError(f"Cannot lock: {obj}.")

    # handle a premature Ctrl+C exit from this context manager
    try:
        signal(SIGTERM, locker._interrupt_handler)
        signal(SIGINT, locker._interrupt_handler)
        # If this is run in a thread, the signal module is not available and raises an exception.
        # ValueError: signal only works in main thread of the main interpreter
        # That's fine; in this case, we don't need to handle signals anyway.
    except ValueError as e:
        _LOGGER.error(f"Failed to set interrupt handler: {e}")

    locker.read_lock()

    try:
        yield obj
    finally:
        locker.read_unlock()


@contextmanager
def write_lock(obj: Union[str, object]) -> object:
    """
    Write-lock file path or object with locker attribute.
    """
    if type(obj) == str:
        locker = ThreeLocker(obj)
    elif hasattr(obj, "locker"):
        locker = obj.locker
    else:
        raise AttributeError(f"Cannot lock: {obj}.")

    # handle a premature Ctrl+C exit from this context manager
    try:
        signal(SIGTERM, locker._interrupt_handler)
        signal(SIGINT, locker._interrupt_handler)
    except ValueError as e:
        _LOGGER.error(f"Failed to set interrupt handler: {e}")

    locker.write_lock()
    try:
        yield obj
    finally:
        locker.write_unlock()


def locked_read_file(filepath, create_file: bool = False) -> str:
    """
    Read a file contents into memory after locking the file. This will prevent
    other ThreeLocker-protected processes from writing to the file while it is
    being read.

    :param str filepath: path to the file that should be read
    """
    if os.path.exists(filepath):
        with read_lock(filepath), open(filepath, "r") as file:
            file_contents = file.read()
    elif create_file:
        _LOGGER.info("File does not exist, but create_file is true. Creating...")
        file_contents = ""
        create_file_racefree(filepath)
    else:
        raise FileNotFoundError(f"No such file: {filepath}")
    return file_contents


def wait_for_locks(lock_paths: Union[List, str], wait_max: int = 10):
    """
    Wait for a lock file to be removed

    :param str lock_paths: path to a file to lock
    :param int wait_max: max wait time if the file in question is already locked
    """
    if not isinstance(lock_paths, list):
        lock_paths = [lock_paths]
    for lock_path in lock_paths:
        wait_for_lock(lock_path, wait_max)


def ensure_write_access(lock_path, strict_ro_locks=False):
    if os.access(os.path.dirname(lock_path), os.W_OK):  # write access
        return True
    else:
        if strict_ro_locks:  # fail; no write access, strict mode
            raise OSError(f"No write access to '{lock_path}'; can't lock file.")
        else:  # warn; no write access, non-strict mode
            _LOGGER.warning(f"No write access to '{lock_path}'; can't lock file.")
            return True


def _create_lock(lock_path, filepath, wait_max) -> None:
    # _LOGGER.debug(f"Creating lock for {filepath} at {lock_path}")
    _LOGGER.debug(f"Creating lock at {os.path.basename(lock_path)}")
    try:
        create_file_racefree(lock_path)
    except FileNotFoundError:
        parent_dir = os.path.dirname(filepath)
        os.makedirs(parent_dir)
        _create_lock(lock_path, filepath, wait_max)
    except Exception as e:
        if e.errno == errno.EEXIST:
            # Rare case: file already exists;
            # the lock has been created in the split second since the
            # last lock existence check,
            # wait for the lock file to be gone, but no longer than
            # `wait_max`.
            _LOGGER.info(
                "The lock has been created in the split second since the "
                "last lock existence check. Waiting"
            )
            wait_for_lock(lock_path, wait_max)
            _create_lock(lock_path, filepath, wait_max)
        else:
            raise e


def _remove_lock(lock_path) -> bool:
    """
    Remove lock

    :param str filepath: path to the file to remove the lock for.
        Not the path to the lock!
    :return bool: whether the lock was found and removed
    """
    _LOGGER.debug(f"Removing lock at {os.path.basename(lock_path)}")
    if os.path.exists(lock_path):
        os.remove(lock_path)
        return True
    return False


def make_all_lock_paths(filepath):
    """
    Create a collection of paths to lock files with given name as base.
    """
    lock_paths = {}
    for type in [READ, WRITE, UNIVERSAL, READ_GLOB]:
        prefix = f"{LOCK_PREFIX}-{type}-" if type else LOCK_PREFIX
        base, name = os.path.split(filepath)
        lock_name = name if name.startswith(prefix) else prefix + name
        lock_paths[type] = lock_name if not base else os.path.join(base, lock_name)
    return lock_paths


def mkabs(path, reldir=None):
    """
    Makes sure a path is absolute; if not already absolute, it's made absolute
    relative to a given directory. Also expands ~ and environment variables for
    kicks.

    :param str path: Path to make absolute
    :param str reldir: Relative directory to make path absolute from if it's
        not already absolute

    :return str: Absolute path
    """

    def xpand(path):
        return os.path.expandvars(os.path.expanduser(path))

    if os.path.isabs(xpand(path)):
        return xpand(path)

    if not reldir:
        return os.path.abspath(xpand(path))

    return os.path.join(xpand(reldir), xpand(path))


# class OneLocker(object):
# def lock(self, type=READ):
#     if not self.filepath:
#         _LOGGER.warning("No filepath, no need to lock.")
#         return True

#     # Check for permissions to write a lock file
#     lock_path = self.lock_paths[type]

#     if not os.access(os.path.dirname(lock_path), os.W_OK):
#         if self.strict_ro_locks:
#             raise OSError(f"No write access to '{lock_path}'; can't lock file.")
#         else:
#             _LOGGER.warning(f"No write access to '{lock_path}'; can't lock file.")
#             self.locked[type] = True
#             return True

#     create_lock(self.filepath, self.wait_max)
#     self.locked[type] = True
#     return True

# def unlock(self, type=READ):
#     if not self.filepath:
#         _LOGGER.warning("No filepath, no need to unlock.")
#         return True

#     # Check for permissions to write a lock file
#     lock_path = make_lock_path(self.filepath)
#     if not os.access(os.path.dirname(lock_path), os.W_OK):
#         if self.strict_ro_locks:
#             raise OSError(f"No write access to '{lock_path}' can't lock file.")
#         else:
#             _LOGGER.warning(f"No write access to '{lock_path}' can't lock file.")
#             self.locked = False
#             return True

#     remove_lock(self.filepath)
#     self.locked = False
#     return True

# def __del__(self):
#     if self.filepath and self.locked:
#         self.unlock()
